#include <gmsh.h>
#include <iostream>
#include <map>
#include <sstream>
#include <Eigen/Dense>
#include <cmath>
#include <iomanip>
#include <stdio.h>
#include <chrono>
#include <omp.h>

// Each element has all its information gathered as in this structure.
struct elementStruct{
    std::size_t tag;            // tag of the element
    double bcValue [2]; // values of [u, u_n]
    size_t nodeTags [2];   // tags of the [first, second] nodes of the element
    size_t midNodeTag;
    double length;
    double x1, y1, x2, y2;
    double midX, midY;
    double deltaX, deltaY;
};

/*
 *  Main
 */

// Fills the "electrostaticPressure" map that gives the electrostatic pressure associated with each (boundary) element tag.
// "nbViews" is the counter of views (useful if other views have to be generated than with this solver). "nodalDisplacements" is
// a map associating the (x,y) displacement to a serie of node tags (if the position of the nodes are not striclty the one
// extracted from the .geo file). "postProcessing" indicates whether the post-processing should be done ('true') or not
// ('false'). "iteration" represents the number of times the solver has been called (some information is printed on the first
// iteration). "untangleMesh" unstangles every BEM domain if set to 'true' (in case the "nodalDisplacements" map contains
// non-zero displacements).
void solverBEM(std::map<int, double> &electrostaticPressure,
               int & nbViews,
               std::map<int,std::pair<double,double>> & nodalDisplacements,
               const bool postProcessing,
               const int iteration,
               const bool untangleMesh);

// Fills the "electrostaticPressure" map. The "nodalDisplacements", "postProcessing" and "iteration" arguments are described in
// the description of the function "solverBEM". "domainName" is used for extracting everything in the appropriate independant BEM
// domain, "phiViewTag" and "electricFieldViewTag" are the view tag of the potential and of the electric field respectively.
// "computePhi" and "computeElectricField" indicates whether the potential and/or the electric field should be computed ('true')
// or not ('false').
void singleDomainSolverBEM(std::map<int, double> &electrostaticPressure,
                           const int nbViews,
                           std::map<int,std::pair<double,double>> & nodalDisplacements,
                           const bool postProcessing,
                           const int iteration,
                           const std::string domainName,
                           const int phiViewTag,
                           const int electricFieldViewTag,
                           const bool computePhi,
                           const bool computeElectricField);

/*
 *  Boundary element
 */

// Fills the "boundaryConditions" map associating to each physical group (a string) the type of boundary condition ('true':
// dirichlet, 'false': neumann) and its value (if any). "domainName" is the name of the current BEM domain.
void getBC(std::map<std::string, std::pair<bool, double>> &boundaryConditions,
           const std::string domainName);

// Fills the "elementVector" vector storing the boundary elements (see "elementStruct" in the 'functions.hpp' file) of the
// current BEM domain, with the appropriate information for each element (the bcValue will be computed later). Also fills the
// "physicalGroups" map associating to each (physical group) name, its tag and dimension. "domainName" is the name of the current
// BEM domain. "nodalDisplacements" is a map associating the (x,y) displacement to a serie of node tags. Returns the number of
// 'dirichlet elements' (useful for the construction/resolution of the system).
std::size_t fillElementVector(std::vector<elementStruct> &elementVector,
                              std::map<std::string, std::pair<int, int>> &physicalGroups,
                              std::map<int, std::pair<double, double>> &nodalDisplacements,
                              const std::string domainName);

// Computes the useful x of a given element and stores it in "element". "nodeIndinces" is a map associating the index of any
// node tag for accessing its coordinates in the "nodeCoords" vector. "nodalDisp1" and "nodalDisp2" are pairs representing the
// potential displacement of the first and second node (respectivelly) of the element. If an element has 3 nodes, the first and
// second nodes are considered to be the extremities of that element.
void computeGeometricParam(elementStruct &element,
                           std::map<std::size_t, std::size_t> &nodeIndices,
                           const std::vector<double> &nodeCoords,
                           const std::pair<double, double> &nodalDisp1,
                           const std::pair<double, double> &nodalDisp2);

// Fills the "A", "B", "c" matrices (/vector) that represent the linear system. "elementVector" is a vector storing the boundary
// elements (see "elementStruct" in the 'functions.hpp' file) of the current BEM domain. "nbDirichletElements" is the number of
// 'dirichlet elements'. "solutionType" indicates whether the elements of the matrices are computed analytically ('true') or
// numerically ('false'). "integrationType" is the type of numerical integration used (in case of numerical computation).
void fillSystem(Eigen::MatrixXd &A,
                Eigen::MatrixXd &B,
                Eigen::MatrixXd &c,
                const std::vector<elementStruct> &elementVector,
                const std::size_t nbDirichletElements,
                const std::string integrationType,
                const bool solutionType);

// Fills the "electrostaticPressure" map associating each (boundary) element tag with its electrostatic pressure. Also sets the
// unknown of each element (either the potential of the electric field) of the "elementVector" vector storing the boundary
// elements (see "elementStruct" in the 'functions.hpp' file) of the current BEM domain. "x" is the solution of the linear
// system. "nbDirichletElements" is the number of 'dirichlet elements', "domainName" is the name of the considered BEM domain.
void computeElectrostaticPressure(std::map<int, double> &electrostaticPressure,
                                  std::vector<elementStruct> &elementVector,
                                  const Eigen::MatrixXd &x,
                                  const std::size_t nbDirichletElements,
                                  const std::string domainName);

/*
 * Post processing
 */

int getModel(std::vector<size_t> &domainNodeTags,
             std::map<std::size_t, std::size_t> &domainNodeIndices,
             std::vector<double> &domainNodeCoords,
             std::vector<int> &domainEntityTags,
             std::map<int, std::size_t> &domainEntityIndices,
             std::vector<std::vector<int>> &domainElementTypes,
             std::vector<std::vector<std::vector<size_t>>> &domainElementTags,
             std::map<std::size_t, std::vector<std::size_t>> &domainElementNodeTags,
             std::map<std::string, std::pair<int, int>> &physicalGroups,
             const std::string domainName);

void computeInternalPhi(std::map<int, double> &internalNodalPhi,
                          const std::vector<double> &domainNodeCoords,
                          const std::vector<size_t> &domainNodeTags,
                          std::map<std::size_t, std::size_t> &domainNodeIndices,
                          const std::vector<elementStruct> &elementVector,
                          const std::string integrationType,
                          std::map<std::size_t, std::vector<std::size_t>> &boundaryNodeToElements,
                          const bool &solutionType);

void computeNodeToElements(std::map<std::size_t, std::vector<std::size_t>> &boundaryNodeToElements,
                           const std::vector<elementStruct> &elementVector);

void boundaryElementTags(std::vector<std::size_t> &domainBoundaryElementTags,
                         const std::vector<std::vector<int>> &domainElementTypes,
                         const std::vector<std::vector<std::vector<std::size_t>>> &domainElementTags,
                         const std::vector<elementStruct> &elementVector,
                         const int domainPhysicalGroupTag);

void phiInitialization(std::vector<std::vector<std::vector<std::vector<double>>>> &phi,
                        std::map<std::size_t, std::vector<std::size_t>> &position,
                        const std::vector<std::vector<int>> &domainElementTypes,
                        const std::vector<std::vector<std::vector<size_t>>> &domainElementTags);

void boundaryElementFinalization(std::vector<std::vector<std::vector<std::vector<double>>>> &phi,
                                 const std::vector<std::vector<std::vector<std::size_t>>> &domainElementTags,
                                 std::map<std::size_t, std::vector<std::size_t>> &domainElementNodeTags,
                                 std::map<int, double> &internalNodalPhi);

void elementNodeData(std::vector<std::vector<std::vector<std::vector<double>>>> &phi,
                     const std::vector<std::vector<std::vector<size_t>>> &domainElementTags,
                     const std::vector<std::vector<int>> &domainElementTypes,
                     std::map<std::size_t, std::vector<std::size_t>> &domainElementNodeTags,
                     const std::vector<elementStruct> &elementVector,
                     std::map<int, double> &internalNodalPhi,
                     const int domainPhysicalGroupTag,
                     std::map<std::size_t, std::vector<std::size_t>> &boundaryNodeToElements);

void computeElementData(std::vector<std::vector<std::vector<std::vector<double>>>> &phi,
                        std::vector<std::vector<std::vector<std::vector<double>>>> &electricField,
                        const std::vector<std::vector<std::vector<std::size_t>>> &domainElementTags,
                        const std::vector<std::vector<int>> &domainElementTypes,
                        const std::vector<int> &domainEntityTags,
                        const std::vector<elementStruct> &elementVector,
                        const bool solutionType,
                        const std::string integrationType,
                        const bool computePhi,
                        const bool computeElectricField);

void displayData(const int viewTag,
                 const std::string modelName,
                 const std::string fileName,
                 const std::string dataType,
                 const std::vector<std::vector<std::vector<size_t>>> &domainElementTags,
                 const std::vector<std::vector<std::vector<std::vector<double>>>> &data,
                 const int viewIndex);

/*
 *  Verification
 */

void displayTime(const double start,
                  const double end,
                  std::string functionName, const bool showTime);

void convergenceFun(const std::vector<int> &domainEntityTags,
                    const std::vector<std::vector<int>> &domainElementTypes,
                    const std::vector<std::vector<std::vector<size_t>>> &domainElementTags,
                    const std::vector<std::vector<std::vector<std::vector<double>>>> &elementalPhi,
                    const std::vector<std::vector<std::vector<std::vector<double>>>> &elementalElectricField);


/*
 *  Computations
 */

double computeAnalyticalH(const double x, const double y, const elementStruct &element);
double computeNumericalG(const double x, const double y, const elementStruct &element, const std::string integrationType);
double computeNumericalGradHx(const double x, const double y, const elementStruct &element, const std::string integrationType);
double computeNumericalGradHy(const double x, const double y, const elementStruct &element, const std::string integrationType);
double computeNumericalGradGx(const double x, const double y, const elementStruct &element, const std::string integrationType);
double computeNumericalGradGy(const double x, const double y, const elementStruct &element, const std::string integrationType);
double computeAnalyticalGradHx(const double x, const double y, const elementStruct &element);
double computeAnalyticalGradHy(const double x, const double y, const elementStruct &element);
double computeAnalyticalGradGx(const double x, const double y, const elementStruct &element);
double computeAnalyticalGradGy(const double x, const double y, const elementStruct &element);
double computeAnalyticalG(const double x, const double y, const elementStruct &element);
